mod common;
mod srp_analyser;
mod formater;
mod file_handler;

use common::*;
use srp_analyser::*;
use formater::*;
use file_handler::*;

use std::path::PathBuf;
use structopt::StructOpt;


#[derive(Debug, StructOpt)]
#[structopt(name = "srp_analysis", about = "Preforms srp analysis.")]
struct Opt {
    /// Approximate preemptions
    #[structopt(short, long)]
    approx: bool,    
    
    /// Output in json format
    #[structopt(short, long)]
    json: bool,    
     
    /// Path to the input json file
    #[structopt(short, long)]
    in_file: Option<PathBuf>,
    
    /// Path to the output file
    #[structopt(short, long)]
    out_file: Option<PathBuf>,
}


fn main() {
    let opt = Opt::from_args();
    //println!("{:?}", opt);

    let tasks: Tasks = match opt.in_file {
        Some(file) => match load_tasks(file) {
            Ok(ts) => ts,
            Err(err) => {
                println!("{}", err);
                return;
            },
        },
        None => create_tasks(),
    };
    
    let (ip, tr) = pre_analysis(&tasks);
    let analysis = analyse(&tasks, &ip, &tr, opt.approx);

    let output: String = match opt.json {
        true => serde_json::to_string(&analysis).unwrap(),
        false => format_analysis(&analysis, &tasks, &ip, &tr),
    };

    match opt.out_file {
        Some(file) => {
            match write_to_file(output, file) {
                Ok(_) => (),
                Err(err) => panic!(err),
            };
        },
        None => println!("{}", output),
    };
}


/*
 * Creates example tasks.
 */
fn create_tasks() -> Tasks { let t1 = Task {
        id: "T1".to_string(),
        prio: 1,
        deadline: 100,
        inter_arrival: 100,
        trace: Trace {
            id: "T1".to_string(),
            start: 0,
            end: 10,
            inner: vec![],
        },
    };

    let t2 = Task {
        id: "T2".to_string(),
        prio: 2,
        deadline: 200,
        inter_arrival: 200,
        trace: Trace {
            id: "T2".to_string(),
            start: 0,
            end: 30,
            inner: vec![
                Trace {
                    id: "R1".to_string(),
                    start: 10,
                    end: 20,
                    inner: vec![Trace {
                        id: "R2".to_string(),
                        start: 12,
                        end: 16,
                        inner: vec![],
                    }],
                },
                Trace {
                    id: "R1".to_string(),
                    start: 22,
                    end: 28,
                    inner: vec![],
                },
            ],
        },
    };

    let t3 = Task {
        id: "T3".to_string(),
        prio: 3,
        deadline: 50,
        inter_arrival: 50,
        trace: Trace {
            id: "T3".to_string(),
            start: 0,
            end: 30,
            inner: vec![Trace {
                id: "R2".to_string(),
                start: 10,
                end: 20,
                inner: vec![],
            }],
        },
    };
    
    // builds a vector of tasks t1, t2, t3
    let tasks: Tasks = vec![t1, t2, t3];
    return tasks;
}

